/*
 * $Id: cilk-conf.h,v 1.1.1.1 1997/02/14 23:07:13 strumpen Exp $
 */
/*
 * $Log: cilk-conf.h,v $
 * Revision 1.1.1.1  1997/02/14 23:07:13  strumpen
 * Create Cilk Repository
 *
 * Revision 1.1.1.1  1997/02/14 22:04:31  strumpen
 * Create cilk repository
 *
 * Revision 1.11  1997/02/05 23:28:40  randall
 * Added SGI support to rts.
 *
 * Revision 1.10  1997/02/02 19:37:20  athena
 * Fixed bug in Solaris/threads.
 *
 * Revision 1.9  1997/01/28 20:25:46  athena
 * Hacked cilk-conf for linux
 *
 * Revision 1.8  1997/01/28 16:39:09  athena
 * Added support for Cilk_alloca()
 *
 * Revision 1.7  1997/01/16 14:23:06  athena
 * Added support for linux
 *
 * Revision 1.6  1996/12/13 14:06:03  athena
 * Added support for Solaris threads
 *
 * Revision 1.5  1996/12/12 23:27:54  athena
 * Cilk2c now understands `register int foo asm("r0")' and the like.
 * Cilk2c understands #pragma segments [+-].
 * Added flexibility to the rts to accomodate the above.
 *
 * Revision 1.4  1996/12/11 20:06:06  athena
 * Simplified management of shared/private segments.  Now Milk
 * manages both segments in the same way, and for both there
 * is the option of using the linker or of manually handling
 * them.
 *
 * Revision 1.3  1996/12/11 17:17:43  athena
 * *** empty log message ***
 *
 * Revision 1.2  1996/12/10 21:41:58  athena
 * Massive renaming of variables to ease port to architechtures
 * whose address space is not so flexible as on Solaris.
 * All global variables are now either SHARED, PRIVATE or SHARED-ReadOnly.
 * All accesses are wrapped with macros, and (almost) all variables
 * are initialized.  The net effect is that it should be easy to
 * put all shared variables in a structure, and things like that.
 *
 * Revision 1.1.1.1  1996/12/03 22:20:56  athena
 * Created repository for Cilk 5 (a.k.a. Milk)
 *
 */

/*
 * Copyright (c) 1996 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */

#ifndef _CILK_CONF_H
#define _CILK_CONF_H

/*
 * Keep track of various statistics
 */
#ifndef CILK_STATS
#define CILK_STATS 1
#endif

/*
 * Keep track of execution time
 */
#ifndef CILK_TIMING
#define CILK_TIMING 1
#endif

/*
 * Measure critical path. EXPENSIVE in terms of execution time
 */
#ifndef CILK_CRITICAL_PATH
#if CILK_TIMING
#define CILK_CRITICAL_PATH 1
#else
#define CILK_CRITICAL_PATH CILK_CP_DISABLED
#endif
#endif

/*
 * Provide the current work and critical path in
 * Cilk_user_work and Cilk_user_critical_path
 */
#define CILK_USER_TIMERS CILK_CRITICAL_PATH

/*
 * Enable rts self-consistency checks.
 */
#ifndef CILK_DEBUG
#define CILK_DEBUG 1
#endif

/*
 * enable support for alloca (slows down common case frame
 * allocation by 20%)
 */
#ifndef CILK_ALLOCA
#define CILK_ALLOCA 1
#endif


#define MEGABYTE (1UL << 20)

/* 
 * Default size of Cilk's heap
 */
#define CILK_DEFAULT_HEAPSIZE (64U * MEGABYTE)
#define CILK_DEFAULT_STACK_DEPTH 32768  

/*
 *  Architecture-specific stuff.
 */

#ifndef CILK_SOLARIS
#define CILK_SOLARIS 0
#endif

#ifndef CILK_SOLARIS_THREADS
#define CILK_SOLARIS_THREADS 0
#endif

#ifndef CILK_LINUX
#define CILK_LINUX 0
#endif

#ifndef CILK_IRIX_THREADS
#define CILK_IRIX_THREADS 0
#endif

#if CILK_SOLARIS
#define CILK_ARCH_H <arch-solaris.h>
#define CILK_MAX_ALIGN sizeof(long long)
#define CILK_MANAGE_SHARED_SEGMENT 0
#define CILK_MANAGE_PRIVATE_SEGMENT 0
#pragma segments +
#define CILK_SEGMENT_ATTRIBUTE(s)  __attribute__((section(s), nocommon))

/*
 * if Cilk manages its own private segment, we need a way to get
 * *one* variable which is really private (maybe in a register
 * if the address space is always shared)
 */
#define CILK_DECLARE_REALLY_PRIVATE(type,id) register type id asm ("g5")
#define CILK_DEFINE_REALLY_PRIVATE(type,id)
#endif

#if CILK_LINUX
#undef linux
/* hack, since cilk2c doesn't understand some syntax in gcc's include files */
extern int __SI__, __QI__, __HI__, __DI__;
#define CILK_ARCH_H <arch-linux.h>
#define CILK_MAX_ALIGN sizeof(long long)
#define CILK_MANAGE_SHARED_SEGMENT 0
#define CILK_MANAGE_PRIVATE_SEGMENT 0
#pragma segments +
#define CILK_SEGMENT_ATTRIBUTE(s)  __attribute__((section(s), nocommon))
#endif

#if CILK_SOLARIS_THREADS
#define CILK_ARCH_H <arch-solaris-threads.h>
#define CILK_MAX_ALIGN sizeof(long long)
#define CILK_MANAGE_SHARED_SEGMENT 0
#define CILK_MANAGE_PRIVATE_SEGMENT 1
#pragma segments -
#define CILK_DECLARE_REALLY_PRIVATE(type,id) register type id asm ("g5")
#define CILK_DEFINE_REALLY_PRIVATE(type,id)
#define CILK_SEGMENT_ATTRIBUTE(s)
#endif

#if CILK_IRIX_THREADS
#define CILK_ARCH_H <arch-irix-threads.h>
#define CILK_MAX_ALIGN sizeof(long long)
#define CILK_MANAGE_SHARED_SEGMENT 0
#define CILK_MANAGE_PRIVATE_SEGMENT 1
#pragma segments -
#define CILK_DECLARE_REALLY_PRIVATE(type,id) register type id asm ("$23")
#define CILK_DEFINE_REALLY_PRIVATE(type,id)
#define CILK_SEGMENT_ATTRIBUTE(s)
#endif

#ifndef CILK_ARCH_H
#error No architecture defined
#endif

/*
 * there should be no need to hack below this line
 */
#if CILK_CRITICAL_PATH && (!CILK_TIMING)
#warning Critical path measurements cannot be enabled because timing is not supported by the rts
#undef CILK_CRITICAL_PATH
#define CILK_CRITICAL_PATH 0
#endif

#define CILK_CHECK(cond, complaints) \
     if (!(cond)) Cilk_die complaints;

#if CILK_DEBUG
#define CILK_ASSERT(ex) \
      (void)((ex) || (Cilk_die(USE_SHARED_RO(assertion_failed_msg), \
             #ex, __LINE__, __FILE__), 0))
#define WHEN_CILK_DEBUG(ex)  ex
#else
#define CILK_ASSERT(ex)
#define WHEN_CILK_DEBUG(ex)
#endif

#if CILK_ALLOCA
#define WHEN_CILK_ALLOCA(a) a
#else
#define WHEN_CILK_ALLOCA(a)
#endif

#if CILK_STATS
#define WHEN_CILK_STATS(a) a
#else
#define WHEN_CILK_STATS(a)
#endif

#if CILK_TIMING
#define WHEN_CILK_TIMING(a) a
#else
#define WHEN_CILK_TIMING(a)
#endif

#if CILK_CRITICAL_PATH
#define WHEN_CILK_CRITICAL_PATH(a) a
#else
#define WHEN_CILK_CRITICAL_PATH(a)
#endif

#if CILK_USER_TIMERS
#define WHEN_CILK_USER_TIMERS(a) a
#else
#define WHEN_CILK_USER_TIMERS(a)
#endif

#define CILK_IM_FREE_MAGIC 0x132DEBFA
#define CILK_CLOSURE_MAGIC 0xDEADBEEF
#define CILK_STACKFRAME_MAGIC 0xBEEF1234

#define CILK_MALLOC_NLISTS 32

/* threshold above which we use slow malloc */
#define CILK_INTERNAL_MALLOC_THRESHOLD 4096
#define CILK_INTERNAL_MALLOC_FREELISTS 9


/*
 * use / don't use the ability of the GNU linker/gcc to
 * handle several memory segments
 */
#ifndef CILK_MANAGE_SHARED_SEGMENT
#define CILK_MANAGE_SHARED_SEGMENT 1
#endif

#ifndef CILK_MANAGE_PRIVATE_SEGMENT
#define CILK_MANAGE_PRIVATE_SEGMENT 0
#endif

#ifndef CILK_SEGMENT_ATTRIBUTE
#error Undefined macro CILK_SEGMENT_ATTRIBUTE
#endif

/*
 * nomenclature of variables:
 *
 * SHARED: read and written by all processes.  Most recent 
 *         value is seen by everybody.
 * PRIVATE: read and written by all processes.  Each process
 *         has a private copy.
 * SHARED_RO: written by the main process, and only read by
 *         children.  It is ok to simply inherit the variable
 *         through fork()
 */

#if CILK_MANAGE_SHARED_SEGMENT

extern struct Cilk_shared_variables *Cilk_shared_vars;
#define DECLARE_SHARED_VARIABLE(before,id,after) 
#define USE_SHARED(id) (Cilk_shared_vars->id)

#else

#define DECLARE_SHARED_VARIABLE(before,id,after) \
    before Cilk_##id##_shared after CILK_SEGMENT_ATTRIBUTE("Cilk_shared")
#define USE_SHARED(id) Cilk_##id##_shared

#endif /* CILK_MANAGE_SHARED_SEGMENT */


#if CILK_MANAGE_PRIVATE_SEGMENT

CILK_DECLARE_REALLY_PRIVATE(struct Cilk_private_variables *, 
			    Cilk_private_vars);
#define DECLARE_PRIVATE_VARIABLE(before,id,after) 
#define USE_PRIVATE(id) (Cilk_private_vars->id)

#else

#define DECLARE_PRIVATE_VARIABLE(before,id,after) \
   before Cilk_##id##_private after CILK_SEGMENT_ATTRIBUTE("Cilk_private")
#define USE_PRIVATE(id) Cilk_##id##_private

#endif /* CILK_MANAGE_PRIVATE_SEGMENT */

/*
 * it would be nice to map these in a segment which is actually
 * read only.
 * I don't provide any option for these -- just using fork
 * is enough.
 */
#define DECLARE_SHARED_RO_VARIABLE(before,id,after) \
  before Cilk_##id##_ro after
#define USE_SHARED_RO(id) Cilk_##id##_ro
#define INIT_SHARED_RO(id, value) USE_SHARED_RO(id) = value

#endif				/* _CILK_CONF_H */




