/*
 * $Id: cilk.h,v 1.2 1997/02/17 23:56:35 randall Exp $
 */
/*
 * $Log: cilk.h,v $
 * Revision 1.2  1997/02/17 23:56:35  randall
 * Small changes.
 *
 * Revision 1.1.1.1  1997/02/14 23:07:14  strumpen
 * Create Cilk Repository
 *
 * Revision 1.1.1.1  1997/02/14 22:04:32  strumpen
 * Create cilk repository
 *
 * Revision 1.15  1997/02/07 21:18:21  athena
 * Added prototype for main()
 *
 * Revision 1.14  1997/02/04 02:53:54  randall
 * Updated header file inclusions.
 *
 * Revision 1.13  1997/02/02 13:52:17  athena
 * Fixed initialization of the lock in Cilk_rand, incorrect under
 * the new locking scheme.  I had to add hooks back in order to do
 * The Right Thing (may hooks be useful for Mingdong's nondeterminator?)
 *
 * Revision 1.12  1997/02/01 22:12:11  athena
 * Reorganized locks.  Now the choice between memory ans OS locks
 * is a command-line option (yes!)
 *
 * Revision 1.11  1997/01/28 18:45:12  athena
 * Minor fixes + testall for alloca()
 *
 * Revision 1.10  1997/01/28 16:39:11  athena
 * Added support for Cilk_alloca()
 *
 * Revision 1.9  1997/01/27 22:34:25  athena
 * Implemented abort
 *
 * Revision 1.8  1996/12/13 19:32:11  athena
 * More minor fixes
 *
 * Revision 1.7  1996/12/12 18:30:49  athena
 * Moved inessential stuff from cilk.h to cilk-internal.h.  Documented
 * protocols.
 *
 * Revision 1.6  1996/12/11 17:17:46  athena
 * *** empty log message ***
 *
 * Revision 1.5  1996/12/10 21:42:05  athena
 * Massive renaming of variables to ease port to architechtures
 * whose address space is not so flexible as on Solaris.
 * All global variables are now either SHARED, PRIVATE or SHARED-ReadOnly.
 * All accesses are wrapped with macros, and (almost) all variables
 * are initialized.  The net effect is that it should be easy to
 * put all shared variables in a structure, and things like that.
 *
 * Revision 1.4  1996/12/10 17:11:53  athena
 * Got rid of the Cilk_result hack --- required modification
 * of cilk2c.  Consequently, added cilk2c to the Cilk-5 repository.
 * Polished Keith's hacks (changed promote into promote_and_detach,
 * and separated the victim part from the thief part).
 *
 * Revision 1.3  1996/12/09 21:57:21  randall
 * Some performance hacks.  Moved inlet data from child closure to
 * list of children in the parent closure.
 *
 * Revision 1.2  1996/12/03 23:09:08  athena
 * Removed all old Cilk-4 process/wall timers (in favor of
 * microstate acocunting)
 *
 * Revision 1.1.1.1  1996/12/03 22:20:56  athena
 * Created repository for Cilk 5 (a.k.a. Milk)
 *
 */

/*
 * Copyright (c) 1996 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */

#ifndef _CILK_H
#define _CILK_H

#include <cilk-conf.h>

/***********************************************************\
 * Architecture-specific definitions and functions
\***********************************************************/

#include CILK_ARCH_H

/***********************************************************\
 * Types used by cilk2c output
\***********************************************************/

/*
 * ProcInfo fields:
 *          |       NORMAL           |               INLET
 * ---------+------------------------+---------------------------------------
 *     size | size of return value   | size of returned (by SPAWN) value
 *    index | index to put ret. val. | index to put ret. (by INLET_CALL) val.
 *    inlet | 0                      | pointer to inlet function
 *  argsize | 0                      | size of inlet's arguments in frame
 * argindex | ???                    | index of inlet's arguments in frame
 *
 * Hack: the field "inlet" for entry 0 contains the address of 
 * the slow function.  The field "size" for entry 0 contains the size
 * of the value returned by the function as a whole (as opposed to the
 * other entries, which contain the size of the value returned by that
 * spawn/sync statement).
 */

typedef struct {
     int size;
     int index;
     void (*inlet) ();
     int argsize;
     int argindex;
} ProcInfo;

/*
 * a linked list of blocks allocated by Cilk_alloca.
 */
struct alloca_header {
     struct alloca_header *next;
     size_t size;
};

/* 
 * a stack frame consists of this header and of a procedure-specific
 * part 
 */
typedef struct StackFrame {
     WHEN_CILK_DEBUG(volatile int magic;)
     int entry;
     int index;			/* 
				 * index outstanding spawn will return to, 
				 * valid only if entry does not specify
				 * it. 
				 */
     ProcInfo *sig;
     WHEN_CILK_ALLOCA(struct alloca_header *alloca_h;)
     WHEN_CILK_TIMING(Cilk_time mycp;)
     WHEN_CILK_TIMING(Cilk_time cp;)
     WHEN_CILK_TIMING(Cilk_time work;)
} StackFrame;

/*
 * a stack is an array of frame pointers 
 */
typedef StackFrame **Stack;

/*
 * The way head and tail work is a bit tricky.  Conceptually,
 * head and tail belong to Closure; however, when a steal
 * occurs, the Closure of a certain queue is replaced by
 * another Closure.  There is no way to initialize the
 * new Closure with the appropriate head and tail, and install
 * it atomically, unless we agree to lock the victim processor
 * (and we don't want to).  Therefore, the closure contains
 * *pointers* to head and tail.  The actual head and tail are allocated
 * on a per-processor basis.  
 *
 * We allocate this little structure to keep the values of head
 * and tail.  It also caches other important quantities used by
 * the cilk2c output, that should reside in the Closure but
 * are cached here for efficiency reasons.
 */
typedef struct ClosureCache {
     volatile StackFrame **head, **tail;
     /* see sched.c for a description of the exception-handling machinery */
     volatile StackFrame **exception;  
     Stack stack;
     CILK_CACHE_LINE_PAD;
} ClosureCache;

/*
 * Functions defined in the architecture-specific file and used
 * in Cilk programs
 */
extern void Cilk_mutex_init(Cilk_mutex *lock);
extern void Cilk_mutex_wait(Cilk_mutex *lock);
extern void Cilk_mutex_signal(Cilk_mutex *lock);
extern int Cilk_mutex_try(Cilk_mutex *lock);
extern void Cilk_dprintf(const char *fmt,...);
extern void Cilk_die(const char *fmt,...);

/*
 * Scheduler functions used in Cilk programs
 */
extern int Cilk_sync(void);
extern int Cilk_exception_handler(void *, int);
extern void Cilk_set_result(void *resultp, int size);
extern void *Cilk_internal_malloc(size_t);
extern void Cilk_internal_free(void *p, size_t size);
extern void Cilk_internal_malloc_final_check(void);
extern void Cilk_after_sync_slow_cp(Cilk_time *work, Cilk_time *cp);
extern void Cilk_abort_standalone(void);
extern void Cilk_abort_slow(void);
extern void Cilk_event_new_thread(void);

/* variables */
DECLARE_PRIVATE_VARIABLE(extern int, _Self, );
#define Self USE_PRIVATE(_Self)

DECLARE_SHARED_RO_VARIABLE(extern size_t, stackdepth, );

DECLARE_SHARED_RO_VARIABLE(extern const char *, assertion_failed_msg ,);

DECLARE_SHARED_VARIABLE(extern Cilk_time, critical_path, );
DECLARE_SHARED_VARIABLE(extern Cilk_time, total_work, );
WHEN_CILK_TIMING(DECLARE_PRIVATE_VARIABLE(extern Cilk_time, last_cp_time,);)
WHEN_CILK_TIMING(DECLARE_PRIVATE_VARIABLE(extern Cilk_time, cp_hack, );)
WHEN_CILK_TIMING(DECLARE_PRIVATE_VARIABLE(extern Cilk_time, work_hack, );)

DECLARE_PRIVATE_VARIABLE(extern Cilk_time, user_work, );
DECLARE_PRIVATE_VARIABLE(extern Cilk_time, user_critical_path ,);
#define Cilk_user_work USE_PRIVATE(user_work)
#define Cilk_user_critical_path USE_PRIVATE(user_critical_path)

/******************************************************
 *  structs containing shared/private variables, etc.
 ******************************************************/
#include <cilk-segments.h>

/***********************************************************\
 * Hacks
\***********************************************************/
#include "hacks.h"

/***********************************************************\
 * Hooks
\***********************************************************/
typedef void (*HookT)(void);

typedef struct hook {
     HookT fn;
     struct hook *next;
} HookList;

extern void Cilk_add_hook(HookList **listp, HookT fn);
extern void Cilk_run_hooks(HookList *list);

#define NULL_HOOK ((HookList *) 0)

/*
 * various types of hooks 
 * This list is incomplete and should be updated as needed
 */
extern HookList *Cilk_init_global_hooks;
extern HookList *Cilk_init_per_process_hooks;

/***********************************************************\
 * Critical path elapsed time calculator
\***********************************************************/

#if CILK_CRITICAL_PATH

DECLARE_PRIVATE_VARIABLE(extern Cilk_time, last_cp_time,);

static inline Cilk_time Cilk_get_elapsed_time(void)
{
     Cilk_time then = USE_PRIVATE(last_cp_time);
     Cilk_time now = Cilk_get_time();

     CILK_ASSERT(now >= then);

     USE_PRIVATE(last_cp_time) = now;
     return now - then;
}

#endif				/* CILK_CRITICAL_PATH */

/* internal malloc stuff, used by cilk2c's output */

/* statistics for Cilk_internal_malloc, used mainly for debugging */
struct Cilk_im_stats {
     int used;         /* bytes used; the sum must be 0 at the end */
     int wasted;       /* wasted because of the way internal_malloc works */
     int allocated;    /* allocated from the central pool */
     CILK_CACHE_LINE_PAD;   /* pad to cache line to avoid false sharing */
};

/*
 * this is written so that it can be inlined and partially
 * evaluated when size is a constant 
 */
#define CILK_INTERNAL_MALLOC_CANONICALIZE_MACRO(size, n) \
     if (size <= n) return n;

static inline int Cilk_internal_malloc_canonicalize(size_t size)
{
     CILK_INTERNAL_MALLOC_CANONICALIZE_MACRO(size, 16);
     CILK_INTERNAL_MALLOC_CANONICALIZE_MACRO(size, 32);
     CILK_INTERNAL_MALLOC_CANONICALIZE_MACRO(size, 64);
     CILK_INTERNAL_MALLOC_CANONICALIZE_MACRO(size, 128);
     CILK_INTERNAL_MALLOC_CANONICALIZE_MACRO(size, 256);
     CILK_INTERNAL_MALLOC_CANONICALIZE_MACRO(size, 512);
     CILK_INTERNAL_MALLOC_CANONICALIZE_MACRO(size, 1024);
     CILK_INTERNAL_MALLOC_CANONICALIZE_MACRO(size, 2048);
     CILK_INTERNAL_MALLOC_CANONICALIZE_MACRO(size, 4096);
     return -1;  /* keep gcc happy */
}

#define CILK_INTERNAL_MALLOC_SIZE_TO_BUCKET(size, n, bucket) \
     if (size <= n) return bucket;

static inline int Cilk_internal_malloc_size_to_bucket(size_t size)
{
     CILK_INTERNAL_MALLOC_SIZE_TO_BUCKET(size, 16, 0);
     CILK_INTERNAL_MALLOC_SIZE_TO_BUCKET(size, 32, 1);
     CILK_INTERNAL_MALLOC_SIZE_TO_BUCKET(size, 64, 2);
     CILK_INTERNAL_MALLOC_SIZE_TO_BUCKET(size, 128, 3);
     CILK_INTERNAL_MALLOC_SIZE_TO_BUCKET(size, 256, 4);
     CILK_INTERNAL_MALLOC_SIZE_TO_BUCKET(size, 512, 5);
     CILK_INTERNAL_MALLOC_SIZE_TO_BUCKET(size, 1024, 6);
     CILK_INTERNAL_MALLOC_SIZE_TO_BUCKET(size, 2048, 7);
     CILK_INTERNAL_MALLOC_SIZE_TO_BUCKET(size, 4096, 8);
     return -1;  /* keep gcc happy */
}

DECLARE_PRIVATE_VARIABLE(extern void *, im_free_lists, [CILK_INTERNAL_MALLOC_FREELISTS]);
DECLARE_SHARED_RO_VARIABLE(extern struct Cilk_im_stats *, im_info, );

static inline void *Cilk_internal_malloc_fast(size_t size)
{
     int bucket;
     void *mem;

     if (size > CILK_INTERNAL_MALLOC_THRESHOLD) 
	  return Cilk_internal_malloc(size);

     bucket = Cilk_internal_malloc_size_to_bucket(size);

     /* look at the free list */
     if ((mem = USE_PRIVATE(im_free_lists)[bucket])) {
	  WHEN_CILK_DEBUG({
	       USE_SHARED_RO(im_info)[Self].used += size;
	  })
	  USE_PRIVATE(im_free_lists)[bucket] = *((void **) mem);
     } else {
	  mem = Cilk_internal_malloc(size);
     }

     return mem;
}

static inline void Cilk_internal_free_fast(void *p, size_t size)
{
     int bucket;

     if (size > CILK_INTERNAL_MALLOC_THRESHOLD) {
	  Cilk_internal_free(p, size);
	  return;
     }

     bucket = Cilk_internal_malloc_size_to_bucket(size);

     WHEN_CILK_DEBUG({
	  USE_SHARED_RO(im_info)[Self].used -= size;
     })

     *((void **)p) = USE_PRIVATE(im_free_lists)[bucket];
     USE_PRIVATE(im_free_lists)[bucket] = p;
}

static inline void Cilk_destroy_frame(StackFrame *f, size_t size)
{
     WHEN_CILK_ALLOCA({
	  struct alloca_header *p;
	  struct alloca_header *q;
	  for (p = f->alloca_h; p; p = q) {
	       q = p->next;
	       Cilk_internal_free(p, p->size);
	  }
     })

     Cilk_internal_free_fast(f, size);
}

static inline void *Cilk_create_frame(size_t size, ProcInfo *sig)
{
     StackFrame *f = Cilk_internal_malloc_fast(size);
     f->sig = sig;
     WHEN_CILK_ALLOCA(f->alloca_h = (struct alloca_header *)0);
     WHEN_CILK_DEBUG(f->magic = CILK_STACKFRAME_MAGIC);

     return (void *)f;
}

/***********************************************************
 *  Options Stuff
 ***********************************************************/

typedef struct
{
  int nproc;            /* number of scheduling threads to run on      */
  int stackdepth;       /* maximum depth of call tree                  */
  size_t heap;          /* heap size in megabytes                      */
  int statlevel;        /* statistics level     (not implemented)      */
  int yieldslice;       /* boolean - yield time slice when stealing    */
  char *infofile_name;  /* filename to print statistics (not impl.)    */
} Cilk_options;

#define CILK_DEFAULT_OPTIONS \
{ \
    1, \
    CILK_DEFAULT_STACK_DEPTH, \
    CILK_DEFAULT_HEAPSIZE, \
    1, \
    1, \
    "-", \
}

/***********************************************************
 *  Cilk2C-only stuff
 ***********************************************************/

/*
 * Library prototypes
 */
void Cilk_init(Cilk_options *options);
void Cilk_exit(void);
#ifdef __CILK2C__
/* we can typecheck cilk procedure pointers correctly */
int Cilk_run(cilk int (*foo)(int), int arg);
#else
/* we can't typecheck cilk procedure pointers */
int Cilk_run(int (*foo)(int), int arg);
#endif

#endif				/* _CILK_H */
