#include <cilk.h>
#include <cilk-lib.h>
#include <cilk-internal.h>
#include <cilk-cilk2c.h>

/* from other parts of the RTS */
void deque_lock(int pn);
void deque_unlock(int pn);
void deque_add_bottom(Closure *cl, int pn);
Closure *deque_xtract_bottom(int pn);
DECLARE_SHARED_VARIABLE(extern int, done,);
void global_init(Cilk_options *options);
void child_main(int id);
static Closure *create_initial_foo_thread(int (*foo)(int), int arg);
DECLARE_SHARED_RO_VARIABLE(extern struct Closure *, invoke_main, );
void Closure_destroy(Closure *t);
Closure *deque_peek_top(int pn);


void Cilk_init(Cilk_options *options)
{
     /* initialize global part of system */
     global_init(options);

     /* initialize the scheduler */
     Cilk_scheduler_init();

     USE_SHARED_RO(invoke_main) = NULL;

     /* create children */
     Cilk_create_children(child_main);
}


void Cilk_exit(void)
{
  USE_SHARED(done) = 1;
}


DECLARE_SHARED_VARIABLE(volatile int, foo_done, );
DECLARE_SHARED_VARIABLE(volatile int, foo_retval, );


int Cilk_run(int (*foo)(int), int arg)
{
  Closure *t;
  int i;
  
  USE_SHARED(foo_done) = 0;
  
  /*
   * make initial closure
   */
  t = create_initial_foo_thread(foo, arg);
  
  /*
   * install initial closure in dummy queue so it will get stolen
   */
  deque_lock(0);
  deque_add_bottom(t, 0);
  deque_unlock(0);
  
  /*
   * Wake up cilk workers
   */
  
  /*
   * wait until everything completes
   */
  while(!USE_SHARED(foo_done))
    ;
  
  /*
   * Put Cilk workers to sleep
   */
  
  /*
   * Make sure all closures are done
   */
  for(i=0; i<USE_SHARED_RO(active_size); i++)
    {
      deque_lock(i);
      CILK_ASSERT(deque_peek_top(i) == NULL);
      deque_unlock(i);
    }
  
  /*
   * Return foo's result
   */
  return USE_SHARED(foo_retval);
}


/************************************************************
 *  Initial thread
 ************************************************************/

/* This is a hand-compiled procedure that calls foo procedure */
struct invoke_foo_frame {
     StackFrame header;
     int (*foo)(int);
     int arg;
     int retval; 
} ;

/* you are not supposed to understand this */

static void invoke_foo_slow(struct invoke_foo_frame *_frame)
{
     int (*foo)(int);
     int arg;
     int tmp;
     Closure *t;

     _START_THREAD_SLOW();
     switch (_frame->header.entry) {
     case 1: 
	  goto _sync1;
     case 2: 
	  goto _sync2;
     }
     
     
     Cilk_user_work = 0;
     Cilk_user_critical_path = 0;
     WHEN_CILK_TIMING(_frame->header.cp = (Cilk_time) 0);
     WHEN_CILK_TIMING(_frame->header.work = (Cilk_time) 0);
     WHEN_CILK_TIMING(_frame->header.mycp = (Cilk_time) 0);
     WHEN_CILK_TIMING(USE_PRIVATE(last_cp_time) = Cilk_get_time());

     foo = _frame->foo;
     arg = _frame->arg;

     _frame->header.entry = 1;
     _BEFORE_SPAWN_SLOW();
     _PUSH_FRAME(_frame);
     tmp = foo(arg);
     _XPOP_FRAME_RESULT(_frame,/* return nothing */,tmp);
     _frame->retval = tmp;
     _AFTER_SPAWN_SLOW();

     if (0) {
     _sync1:
	  ;
     }
     _AT_THREAD_BOUNDARY_SLOW();
     
     _BEFORE_SYNC_SLOW();
     if (_SYNC) {
          _frame->header.entry=2;
	  return;
     _sync2:
	  ;
     }
     _AFTER_SYNC_SLOW();
     _AT_THREAD_BOUNDARY_SLOW();

     WHEN_CILK_TIMING(USE_SHARED(critical_path) = _frame->header.mycp);
     WHEN_CILK_TIMING(USE_SHARED(total_work) = _frame->header.work);
     
     USE_SHARED(foo_retval) = _frame->retval;
     
     deque_lock(Self);
     t = deque_xtract_bottom(Self);
     CILK_ASSERT(t->frame == &_frame->header);
     deque_unlock(Self);
     Cilk_free(_frame);
     Closure_destroy(t);
     
     USE_SHARED(foo_done) = 1;

     return;
}


static ProcInfo invoke_foo_sig[]={
     {sizeof(int), sizeof(struct invoke_foo_frame ), invoke_foo_slow, 0, 0},
     {sizeof(int), offsetof(struct invoke_foo_frame, retval), 0, 0, 0},
     {0, 0, 0, 0, 0}
};


static Closure *create_initial_foo_thread(int (*foo)(int), int arg)
{
     Closure *t;
     struct invoke_foo_frame *f;

     /* create a frame for invoke_foo */
     t = Cilk_Closure_create();
     t->parent = (Closure *) NULL;
     t->status = CLOSURE_READY;

     f = Cilk_malloc(sizeof(struct invoke_foo_frame));
     f->header.entry = 0;
     f->header.sig = invoke_foo_sig;
     WHEN_CILK_DEBUG(f->header.magic = CILK_STACKFRAME_MAGIC);

     f->foo = foo;
     f->arg = arg;

     t->frame = &f->header;
     return t;
}
